package Image;
use Fcntl qw(:flock);
use strict;
use bytes;
######################################################################
# Web画像分析/GIF結合モジュール
# Image.pm 1.3.2
# Copyright (C) 1999-2018 CMONOS Co. Ltd.
#
# support@cmonos.jp
# http://cmonos.jp/
# 最終更新日：2012-09-01
######################################################################
#
# このサブルーチン集の使い方
# ---------------------------------------------------------------------------------------
#	(0) オブジェクトを作る。
#		require CMONOS::CGI_Lib::Image;
#		obj = Image->new;
#
# ---------------------------------------------------------------------------------------
#	(1) (width,height,extension,mime_type) = obj->info(\image[,maxwidth,maxheight]);
#		imageの情報を取得します。
# 			width			画像の横幅（ピクセル）
# 			height			画像の高さ（ピクセル）
# 			extension		拡張子（jpg/gif/png以外の場合空）
# 			mime_type		MIME_TYPE
#
# 			image			画像ファイルのパス/画像ファイルのデータを収めた変数のリファレンス
# 							例）パス指定：@info = Stat($path);　変数指定：@info = Stat(\$data);
# 			maxwidth		画像の最大横幅（ピクセル）
# 			maxheight		画像の最大高さ（ピクセル）
# 							※ 最大値を超えていると画像情報を返しません。
#
# ---------------------------------------------------------------------------------------
#	(2) (width,height,extension,mime_type) = obj->input_info(input_obj,name,no,maxwidth,maxheight);
#		アップロードされた画像の情報を取得します。
#		また、MacBinaryを削除し適切なMIME_TYPEを指定します。
# 			width			画像の横幅（ピクセル）
# 			height			画像の高さ（ピクセル）
# 			extension		拡張子（.jpg/.gif/.png以外の場合空）
# 			mime_type		MIME_TYPE
#
# 			input_obj		アップロードされたデータが格納されているオブジェクト
# 							※ obj->decode;で使ったオブジェクトを指定して下さい。
# 			name			画像アップロードフォームの名前
# 			no				同一名フォームが複数ある場合のフォーム番号（デフォルト=0）
# 			maxwidth		画像の最大横幅（ピクセル）
# 			maxheight		画像の最大高さ（ピクセル）
#
#	(2)-2-1 obj->is_not_a_web_image(input_obj,name,no,maxwidth,maxheight,ext,force);
#		アップロードされたファイルがweb用画像でないとき1を返します。
#		また、MacBinaryを削除し適切なMIME_TYPEを設定します。
#
#	(2)-2-2 obj->is_not_web_image(input_obj,maxwidth,maxheight,ext,force);
#		アップロードされたファイルの中にweb用画像でないものが含まれるとき1を返します。
#		また、MacBinaryを削除し適切なMIME_TYPEを設定します。
# 			input_obj		アップロードされたデータが格納されているオブジェクト
# 							※ obj->decode;で使ったオブジェクトを指定して下さい。
# 			maxwidth		画像の最大横幅（ピクセル）
# 			maxheight		画像の最大高さ（ピクセル）
# 			ext				画像形式の限定（jpg/gif/png のいずれかを指定します）
#			force			入力の強制
#
# ---------------------------------------------------------------------------------------
#	(3)-1-0 (width,height) = obj->image_resize(o_width,o_height,ratio);
#		指定された割合で拡大縮小した画像サイズを返します。
# 			width			リサイズ後の画像の横幅（ピクセル）
# 			height			リサイズ後の画像の高さ（ピクセル）
#
# 			o_width			オリジナルの画像の横幅（ピクセル）
# 			o_height		オリジナルの画像の高さ（ピクセル）
# 			ratio			リサイズ倍率
#
#	(3)-1-1 (width,height,o_width,o_height) = obj->resize(image,ratio);
#		指定された割合で拡大縮小した画像サイズを返します。
# 			width			リサイズ後の画像の横幅（ピクセル）
# 			height			リサイズ後の画像の高さ（ピクセル）
# 			o_width			オリジナルの画像の横幅（ピクセル）
# 			o_height		オリジナルの画像の高さ（ピクセル）
# 			extension		拡張子（jpg/gif/png以外の場合空）
# 			mime_type		MIME_TYPE
#
# 			image			画像ファイルのパス/画像ファイルのデータを収めた変数のリファレンス
# 			ratio			リサイズ倍率
#
#	(3)-2-0 (width,height) = obj->image_fit(o_width,o_height,maxwidth,maxheight);
#		指定された最大値を超えないよう縮小した画像サイズを返します。
# 			width			リサイズ後の画像の横幅（ピクセル）
# 			height			リサイズ後の画像の高さ（ピクセル）
#
# 			o_width			オリジナルの画像の横幅（ピクセル）
# 			o_height		オリジナルの画像の高さ（ピクセル）
# 			maxwidth		画像の最大横幅（ピクセル）／0で制限なし
# 			maxheight		画像の最大高さ（ピクセル）／0で制限なし
#
#	(3)-2-1 (width,height,o_width,o_height) = obj->fit(image,maxwidth,maxheight);
#		指定された最大値を超えないよう縮小した画像サイズを返します。
# 			width			リサイズ後の画像の横幅（ピクセル）
# 			height			リサイズ後の画像の高さ（ピクセル）
# 			o_width			オリジナルの画像の横幅（ピクセル）
# 			o_height		オリジナルの画像の高さ（ピクセル）
# 			extension		拡張子（.jpg/.gif/.png以外の場合空）
# 			mime_type		MIME_TYPE
#
# 			image			画像ファイルのパス/画像ファイルのデータを収めた変数のリファレンス
# 			maxwidth		画像の最大横幅（ピクセル）／0で制限なし
# 			maxheight		画像の最大高さ（ピクセル）／0で制限なし
#
# ---------------------------------------------------------------------------------------
#	(4)-0 obj->image_tag(image,o_width,o_height[,border,alt,option,maxwidth,maxheight[,base,base_option[,viewer_disabled]]]);
#		指定された画像の<img />タグを返します。
# 			image			画像ファイルのパス（URLも可）
# 			o_width			オリジナルの画像の横幅（ピクセル）
# 			o_height		オリジナルの画像の高さ（ピクセル）
# 			border			枠線幅
# 			alt				ALT
# 			option			IMGタグのオプション
# 			maxwidth		画像の最大横幅（ピクセル）／0で制限なし
# 			maxheight		画像の最大高さ（ピクセル）／0で制限なし
# 			base			タグを書き込むファイルのパス
# 			base_option		abs=画像のパスを絶対パスにする
# 			viewer_disabled	1=実寸画像を新しいウィンドウで開かない
#
#	(4)-1 obj->tag(image[,border,alt,option,maxwidth,maxheight[,base,base_option]]);
#		指定された画像の<img />タグを返します。
# 			image			画像ファイルのパス（必ず相対パスで指定）
# 			border			枠線幅
# 			alt				ALT
# 			option			IMGタグのオプション
# 			maxwidth		画像の最大横幅（ピクセル）／0で制限なし
# 			maxheight		画像の最大高さ（ピクセル）／0で制限なし
# 			base			タグを書き込むファイルのパス
# 			base_option		abs=画像のパスを絶対パスにする
#
# ---------------------------------------------------------------------------------------
#	(5) (width,height) = obj->gif_cat(\@images[,out,vector,@color]);
#		GIF画像を結合します。
#		※アニメーションGIFは一度に表示されず、オリジナル画像ごとに順番に表示されます。
# 			width			結合後の画像の横幅（ピクセル）
# 			height			結合後の画像の高さ（ピクセル）
#
# 			@images			結合するGIFファイルのパス（"\n"で改行/改列）
# 			out				結合した画像のパス（設定なしの場合STDOUTに出力）
# 			vector			結合する方向（"h"=水平方向/"v"=垂直方向）
# 			color			色変更設定（ChangeColorTableからimage,outを省略）
# 							使用例）@images = ('1.gif','2.gif',"\n",'3.gif','4.gif');
# 									$image->gif_cat(\@images,'','h');
# 							　　　　'1.gif','2.gif'
# 							　　　　'3.gif','4.gif'　←結合される配置
#
#	(5)-0 obj->set_cookie(limit,cname,cookie,domain,path);
#		結合した画像を標準出力する際、発行するCookieを設定します。
#		このcookieはlimit日後に消去されます。
#		obj->{'Set_Cookie'} にSet-Cookieヘッダを格納。
#
#		limit		Cookie期限（時間単位）／設定無し=そのセッションのみで消える）
#		cname		Cookieの名前
#		cookie		Cookieに記憶させるデータ
#		domain		Cookieを設定するドメイン
#		path		Cookieを設定するパス（相対パスでもOK）
#
# ---------------------------------------------------------------------------------------
#	(6)-1 (width,height) = obj->change_color_table(image,out,'slide',color[,option[,rev]]);
#		GIF画像の配色表を指定された色方向にずらせます。
# 			width			画像の横幅（ピクセル）
# 			height			画像の高さ（ピクセル）
#
# 			image			オリジナル画像のパス/画像データを収めた変数のリファレンス
# 			out				配色表を変更した画像の出力先
# 			color			変換基準色（例："+666666"）
# 			abs				1=変換後の色が0-255に収まらないとき絶対値を取る
# 			rev				補正色
#
# 			例）	補色に変換：obj->change_color_table(image,out,'slide','+ffffff','abs');
# 					赤色のみ残す：obj->change_color_table(image,out,'slide','-00ffff');
# 					-灰色+黄色：obj->change_color_table(image,out,'slide','+ffff00','','-333333');
#
#	(6)-2 (width,height) = obj->change_color_table(image,out,'change',@colors);
#		GIF画像の配色表を入れ換えます
# 			image			オリジナル画像のパス/画像データを収めた変数のリファレンス
# 			out				配色表を変更した画像の出力先
# 			@colors			新しい配色表
#
# 			例）	白黒に変換：obj->change_color_table(image,out,'change','000000','ffffff');
#
#	(6)-3 obj->change_color_table_multi(dir,out,type,@setting);
#		ディレクトリの中にあるGIF画像の配色表を一括変換します。エラーがなければ1を返します。
# 			dir				オリジナル画像のあるディレクトリのパス
# 			out				変換した画像を収めるディレクトリのパス
#
# 			※	この他の設定は(6)-1/2に準じます。
#
######################################################################
#	■参考資料
#	このスクリプトを作成するに当たり以下のサイトを参考にしました。
#	有益な情報を公開して下さっている諸兄に、心から感謝いたします。
#	http://www.geocities.co.jp/SiliconValley/3453/gif_info/index_jp.html
#	http://www.venus.dti.ne.jp/~yosinobu/filefmts/JFIF_JJ.TXT
#	http://www01.u-page.so-net.ne.jp/xb3/y-koujin/png_fmt/jp/Tr/png.html
#	http://www.bekkoame.ne.jp/~siisise/format.html#JPEG
######################################################################
#	■利用規約
#	このスクリプトの著作権はCMONOSが保有しています。
#	このサブルーチン集は、以下の条件に従うかぎり自由に再配布できます。
#		・著作権表示を変更してはいけません。
#		・修正を加えた場合は、修正者名・修正者の連絡先・修正個所・修正日
#		　時を明記して下さい。
#	このスクリプトを利用したことによって生じた、いかなる損害にも
#	CMONOSは責任を負いません。
######################################################################
#	(0) オブジェクトを作る。
sub new ($) {
    my $pkg  = shift;
    my $this = {};
    bless $this;
    $this;
}

sub image_magick_enabled ($) {
    my $this = shift;
    if ( !defined $cgi::image_magick_enabled ) {
        if ( eval 'use Image::Magick;1;' ) {
            $cgi::image_magick_enabled = 1;
        }
        else {
            $cgi::image_magick_enabled = 0;
        }
    }
    return $cgi::image_magick_enabled;
}

sub set_clipboard ($;$) {
    my ( $this, $path ) = @_;
    $this->{'clipboard'} = $path if $path;
}

######################################################################
#	(1)-0-0 (height,width,extension,mime_type) = obj->info(\image);
#		imageの情報を取得します。
sub info ($$;$$$$$) {
    my ( $this, $image, $maxwidth, $maxheight, $skip_max, $is_thumbnail,
        $max_size )
        = @_;
    my ( $header, $pos, @info );
    local (*IMAGE);
    for ( $pos = 0; $pos <= 128; $pos += 128 ) {
        my $image_handle = $this->open_image($image);
        unless ( ref $image_handle ) {
            push(
                @{ $this->{'dialog'} },
                $this->decode_utf8(
                    '指定された画像を読み込めませんでした。'
                )
                )
                unless grep( $_ eq $this->decode_utf8(
                    '指定された画像を読み込めませんでした。'
                ),
                @{ $this->{'dialog'} } );
            return ();
        }
        $header = $this->read_image( $image_handle, $pos, 24 );
        $this->close_image($image_handle);
        if (   ( @info = $this->gif_info( $image, $header ) ) > 1
            || ( @info = $this->png_info( $image, $header ) ) > 1
            || ( @info = $this->jpg_info( $image, $header ) ) > 1 )
        {
            last;
        }
    }
    if ( ( $is_thumbnail || $maxwidth || $maxheight ) && $info[4] ) {
        $this->fix_orientation( $image, $info[4] );
        ( $info[0], $info[1], $info[2], $info[3] ) = $this->jpg_info($image);
    }
    if ( !defined $info[0] || !defined $info[1] || !$info[0] || !$info[1] ) {
        push(
            @{ $this->{'dialog'} },
            $this->decode_utf8('未知の画像フォーマットです。')
        );
        return ();
    }
    elsif ( !$info[2] || $info[2] !~ /^(gif|jpg|png)$/ ) {
        push(
            @{ $this->{'dialog'} },
            $this->decode_utf8(
                'Web用の画像ファイルではありません。')
        );
        return ();
    }
    elsif (
        (      defined $maxwidth
            && $maxwidth
            && $maxwidth > 0
            && $info[0] > $maxwidth
        )
        || (   defined $maxheight
            && $maxheight
            && $maxheight > 0
            && $info[1] > $maxheight )
        )
    {
        my ( $width, $height )
            = $this->image_fit( $info[0], $info[1], $maxwidth, $maxheight );
        if ($this->image_file_resize(
                $image, $width, $height, $is_thumbnail, $info[4]
            )
            )
        {
            $info[0] = $width;
            $info[1] = $height;
            undef $is_thumbnail;
        }
        elsif ( !$skip_max ) {
            push(
                @{ $this->{'dialog'} },
                $this->decode_utf8(
                          '画像の横幅を'
                        . $maxheight
                        . 'ピクセル以下にしてください。'
                )
            ) if $maxwidth > 0 && $info[0] > $maxwidth;
            push(
                @{ $this->{'dialog'} },
                $this->decode_utf8(
                          '画像の高さを'
                        . $maxheight
                        . 'ピクセル以下にしてください。'
                )
            ) if $maxheight > 0 && $info[1] > $maxheight;
            return ();
        }
    }
    $this->clear_all_profiles($image)
        if $is_thumbnail
        || (
        $max_size
        && (   ( !ref $image && $max_size < -s $image )
            || ( ref $image && $max_size < length $$image ) )
        );
    return @info;
}

sub get_orientation($$;$) {
    my ( $this, $image, $exif ) = @_;
    if ( !$exif ) {
        my (@info) = $this->jpg_info($image);
        $exif = $info[4];
    }
    return ( !$exif || !$exif->{'274'} || $exif->{'274'} <= 2 )
        ? 1
        : $exif->{'274'};
}

sub fix_orientation($$;$) {
    my ( $this, $image, $exif ) = @_;
    return 0 if !$this->image_magick_enabled;
    my $orientation = $this->get_orientation( $image, $exif );
    return 1 if !grep( $orientation == $_, 3, 6, 8 );
    my ($path);
    if ( ref $image ) {
        $this->{'clipboard'} = $env::clipboard
            if !$this->{'clipboard'} && $env::clipboard;
        return 0 if !$this->{'clipboard'};
        local (*IMAGE);
        $path = $this->{'clipboard'} . '/' . $this->temp_file_name;
        if ( open( IMAGE, ">>$path" ) ) {
            binmode IMAGE;
            flock( IMAGE, LOCK_EX );
            truncate( IMAGE, 0 );
            seek( IMAGE, 0, 0 );
            print IMAGE $$image;
            close(IMAGE);
        }
        else {
            return 0;
        }
    }
    else {
        $path = $image;
    }
    my $temp   = $path . '.temp';
    my $magick = Image::Magick->new;
    $magick->Read($path);
    $magick->Rotate(
        degrees => (
            ( $orientation == 3 ) ? 180 : ( $orientation == 6 ) ? 90 : 270
        )
    );
    $magick->Profile( name => "*", profile => "" );
    $magick->Write($temp);
    undef $magick;

    if ( -s $temp ) {
        rename( $temp, $path );
    }
    else {
        unlink $temp;
        unlink $path;
        return 0;
    }
    if ( ref $image ) {
        local (*IMAGE);
        if ( open( IMAGE, $path ) ) {
            binmode IMAGE;
            flock( IMAGE, LOCK_EX );
            $$image = join( '', <IMAGE> );
            close(IMAGE);
            unlink $path;
        }
        else {
            return 0;
        }
    }
    return 1;
}

sub image_file_resize ($$$$;$) {
    my ( $this, $image, $width, $height, $is_thumbnail ) = @_;
    return 0 if !$this->image_magick_enabled;
    my ($path);
    if ( ref $image ) {
        $this->{'clipboard'} = $env::clipboard
            if !$this->{'clipboard'} && $env::clipboard;
        return 0 if !$this->{'clipboard'};
        local (*IMAGE);
        $path = $this->{'clipboard'} . '/' . $this->temp_file_name;
        if ( open( IMAGE, ">>$path" ) ) {
            binmode IMAGE;
            flock( IMAGE, LOCK_EX );
            truncate( IMAGE, 0 );
            seek( IMAGE, 0, 0 );
            print IMAGE $$image;
            close(IMAGE);
        }
        else {
            return 0;
        }
    }
    else {
        $path = $image;
    }
    my $temp   = $path . '.temp';
    my $magick = Image::Magick->new;
    $magick->Read($path);
    if ($is_thumbnail) {
        $magick->Thumbnail( 'width' => $width, 'height' => $height );
    }
    else {
        $magick->Resize( 'width' => $width, 'height' => $height );
    }
    $magick->Write($temp);
    undef $magick;
    if ( -s $temp ) {
        rename( $temp, $path );
    }
    else {
        unlink $temp;
        unlink $path;
        return 0;
    }
    if ( ref $image ) {
        local (*IMAGE);
        if ( open( IMAGE, $path ) ) {
            binmode IMAGE;
            flock( IMAGE, LOCK_EX );
            $$image = join( '', <IMAGE> );
            close(IMAGE);
            unlink $path;
        }
        else {
            return 0;
        }
    }
    return 1;
}

sub clear_all_profiles ($$) {
    my ( $this, $image ) = @_;
    return 0 if !$this->image_magick_enabled;
    my ($path);
    if ( ref $image ) {
        $this->{'clipboard'} = $env::clipboard
            if !$this->{'clipboard'} && $env::clipboard;
        return 0 if !$this->{'clipboard'};
        local (*IMAGE);
        $path = $this->{'clipboard'} . '/' . $this->temp_file_name;
        if ( open( IMAGE, ">>$path" ) ) {
            binmode IMAGE;
            flock( IMAGE, LOCK_EX );
            truncate( IMAGE, 0 );
            seek( IMAGE, 0, 0 );
            print IMAGE $$image;
            close(IMAGE);
        }
        else {
            return 0;
        }
    }
    else {
        $path = $image;
    }
    my $temp   = $path . '.temp';
    my $magick = Image::Magick->new;
    $magick->Read($path);
    $magick->Profile( name => "*", profile => "" );
    $magick->Write($temp);
    undef $magick;
    if ( -s $temp ) {
        rename( $temp, $path );
    }
    else {
        unlink $temp;
        unlink $path;
        return 0;
    }
    if ( ref $image ) {
        local (*IMAGE);
        if ( open( IMAGE, $path ) ) {
            binmode IMAGE;
            flock( IMAGE, LOCK_EX );
            $$image = join( '', <IMAGE> );
            close(IMAGE);
            unlink $path;
        }
        else {
            return 0;
        }
    }
    return 1;
}

sub temp_file_name ($) {
    my $this = shift;
    return $ENV{'UNIQUE_ID'} . ++$Image::unique_number if $ENV{'UNIQUE_ID'};
    my $name
        = ( ($default::id) ? $default::id . '_' : '' )
        . time
        . $$
        . $ENV{'REMOTE_ADDR'}
        . ++$Image::unique_number;
    $name =~ s/[^0-9a-zA-Z_]//g;
    return $name;
}

#---------------------------------------------------------------------
#	(1)-0-1 画像を開く
sub open_image {
    my ( $this, $image ) = @_;
    if ( ref $image ) {
        utf8::encode($$image) if utf8::is_utf8($$image);
        return $image;
    }
    elsif ( $image && open( IMAGE, $image ) ) {
        binmode IMAGE;
        flock( IMAGE, LOCK_EX );
        return \*IMAGE;
    }
    else {
        return;
    }
}

#---------------------------------------------------------------------
#	(1)-0-2 データの取得
sub read_image {
    my ( $this, $image, $pos, $buf ) = @_;
    my $out;
    if ( ref $image eq 'GLOB' ) {
        seek( $image, $pos, 0 );
        read( $image, $out, $buf );
    }
    else {
        $out = bytes::substr( $$image, $pos, $buf );
    }
    return $out;
}

#---------------------------------------------------------------------
#	(1)-0-2 画像を閉じる
sub close_image {
    my ( $this, $image ) = @_;
    close($image) if ref $image eq 'GLOB';
}

#---------------------------------------------------------------------
#	(1)-1-1 (width,height,extension,mime_type) = obj->gif_info(image);
#		GIF画像の情報を取得します。
sub gif_info {
    my ( $this, $image, $header ) = @_;
    my ( $width, $height );
    local (*IMAGE);
    if ( !defined $header || $header eq '' ) {
        $image = $this->open_image($image);
        if ( !defined $image || $image eq '' ) {
            push(
                @{ $this->{'dialog'} },
                $this->decode_utf8(
                    '画像を読み込めませんでした。')
            );
            return ();
        }
        $header = $this->read_image( $image, 0, 10 );
        $this->close_image($image);
    }
    if ( $header =~ /^\x47\x49\x46/ ) {
        $width  = unpack( 'v', bytes::substr( $header, 6, 2 ) );
        $height = unpack( 'v', bytes::substr( $header, 8, 2 ) );
        return ( $width, $height, 'gif', 'image/gif' );
    }
    else {
        return ();
    }
}

#---------------------------------------------------------------------
#	(1)-1-2 (width,height,extension,mime_type) = obj->png_info(image);
#		PNG画像の情報を取得します。
sub png_info {
    my ( $this, $image, $header ) = @_;
    my ( $width, $height );
    local (*IMAGE);
    if ( !defined $header || $header eq '' ) {
        $image = $this->open_image($image);
        if ( !defined $image || $image eq '' ) {
            push(
                @{ $this->{'dialog'} },
                $this->decode_utf8(
                    '画像を読み込めませんでした。')
            );
            return ();
        }
        $header = $this->read_image( $image, 0, 24 );
        $this->close_image($image);
    }
    if ( $header =~ /^\x89\x50\x4e\x47\x0d\x0a\x1a\x0a/ ) {
        $width  = unpack( 'N', bytes::substr( $header, 16, 4 ) );
        $height = unpack( 'N', bytes::substr( $header, 20, 4 ) );
        return ( $width, $height, 'png', 'image/png' );
    }
    else {
        return ();
    }
}

#---------------------------------------------------------------------
#	(1)-1-3 (width,height,extension,mime_type) = obj->jpg_info(image);
#		JPEG画像の情報を取得します。
sub jpg_info {
    my ( $this, $image, $header, $required_tags ) = @_;
    my ( $pos, $width, $height, $exif );
    $required_tags = [
        270,      #imageDescription
        271,      #make
        272,      #model
        274,      #Orientation
        306,      #dateTime
        315,      #artist
        33432,    #copyright
        33434,    #exposureTime
        33437,    #fNumber
        37386,    #focalLength
        37396,    #subjectArea
        41986,    #exposureMode
        41987,    #whiteBalance
        41988,    #digitalZoomRatio
        41989,    #focalLengthIn35mmFilm
        41990,    #sceneCaptureType
        41991,    #gainControl
        41992,    #contrast
        41993,    #saturation
        41994,    #sharpness
        34665,    #Exif
        34853, 1, 2, 3, 4, 5, 6, 7    #GPS
    ] if !ref $required_tags;
    local (*IMAGE);
    $image = $this->open_image($image);
    if ( !defined $image || $image eq '' ) {
        push(
            @{ $this->{'dialog'} },
            $this->decode_utf8('画像を読み込めませんでした。')
        );
        return ();
    }
    $header = $this->read_image( $image, 0, 11 )
        if !defined $header || $header eq '';
    if ( $header !~ /^\xff\xd8\xff/ ) {
        $this->close_image($image);
        return ('error');
    }
    $pos = 2;
    while (1) {
        my $marker = $this->read_image( $image, $pos, 2 );
        $pos += 2;
        my $length = unpack( 'n', $this->read_image( $image, $pos, 2 ) );
        if ( !$marker || $marker !~ /^\xff/ || $marker =~ /[\xd9\xda]$/ ) {
            last;
        }
        elsif ( $marker =~ /\xe1$/ ) {
            if (unpack( 'A6', $this->read_image( $image, $pos + 2, 6 ) ) eq
                "Exif" )
            {
                $exif
                    = $this->exif(
                    $this->read_image( $image, $pos + 8, $length - 8 ),
                    $required_tags );
            }
        }
        elsif ( $marker =~ /[\xc0-\xc3]$/ ) {
            $height = unpack( 'n', $this->read_image( $image, $pos + 3, 2 ) );
            $width  = unpack( 'n', $this->read_image( $image, $pos + 5, 2 ) );
            last;
        }
        $pos += $length;
    }
    $this->close_image($image);
    return ( $width && $height )
        ? ( $width, $height, 'jpg', 'image/jpeg', $exif )
        : ();
}

# http://www.kanzaki.com/docs/sw/exif-extract.html
sub exif ($$;$) {
    my ( $this, $data, $required_tags ) = @_;
    my $exif = {};
    my ( $long_tmpl, $short_tmpl )
        = ( unpack( "A2", substr( $data, 0, 2 ) ) eq 'MM' )
        ? ( "N", "n" )
        : ( "V", "v" );
    my $offset = unpack( $long_tmpl, substr( $data, 4, 4 ) );
    my $ifd_num = 0;
    1 while (
        $offset = $this->read_ifd(
            \$data,     $offset,     0,              $exif,
            $long_tmpl, $short_tmpl, $required_tags, $ifd_num++
        )
    );
    return $exif;
}

# IFDの解析
sub read_ifd ($$$$$$$;$$) {
    my ($this,       $data_ref,      $offset,
        $parent_tag, $exif,          $long_tmpl,
        $short_tmpl, $required_tags, $ifd_num
    ) = @_;
    my $num_dir_entry
        = unpack( $short_tmpl, substr( $$data_ref, $offset, 2 ) );
    my $next_offset = unpack( $long_tmpl,
        substr( $$data_ref, $offset + 2 + $num_dir_entry * 12, 4 ) );
    $offset += 2;
    while ($num_dir_entry) {
        $this->read_dir_entry(
            $data_ref,        substr( $$data_ref, $offset, 12 ),
            --$num_dir_entry, $parent_tag,
            $exif,            $long_tmpl,
            $short_tmpl,      $required_tags,
            $ifd_num
        );
        $offset += 12;
    }
    return $next_offset;
}

# ディレクトリエントリの解析
sub read_dir_entry ($$$$$$$$;$$) {
    my ($this,          $data_ref, $dir_entry, $num_dir_entry,
        $parent_tag,    $exif,     $long_tmpl, $short_tmpl,
        $required_tags, $ifd_num
    ) = @_;
    my ( $tag, $type, $num_data, $dir_offset )
        = unpack( "$short_tmpl$short_tmpl$long_tmpl$long_tmpl", $dir_entry );
    my $data_value = '';
    my $dir_size = ( 1, 1, 2, 4, 8 )[ $type - 1 ] || 1;
    if (   !ref $required_tags
        || !@$required_tags
        || grep( $tag == $_, @$required_tags ) )
    {
        if ( $tag == 34665 || $tag == 34853 || $tag == 40965 ) {
            1 while $dir_offset
                = $this->read_ifd( $data_ref, $dir_offset, $tag, $exif,
                $long_tmpl, $short_tmpl, $required_tags );
        }
        else {
            $type = 2 if $tag == 37510;
            if ( $num_data * $dir_size > 4 ) {
                foreach my $i ( 0 .. $num_data - 1 ) {
                    $data_value
                        .= ( ( $type == 5 && $data_value ) ? ' ' : '' )
                        . $this->read_exif_data(
                        $type,
                        substr(
                            $$data_ref, $dir_offset + $i * $dir_size,
                            $dir_size
                        ),
                        $long_tmpl,
                        $short_tmpl
                        );
                }
            }
            else {
                foreach my $i ( 0 .. $num_data - 1 ) {
                    $data_value
                        .= ( ( $type == 5 && $data_value ) ? ' ' : '' )
                        . $this->read_exif_data( $type,
                        substr( $dir_entry, 8 + $i * $dir_size, $dir_size ),
                        $long_tmpl, $short_tmpl );
                }
            }
            $exif->{ '999999.' . $ifd_num } = {}
                if $ifd_num && !ref $exif->{ '999999.' . $ifd_num };
            my $exif_hash
                = ($ifd_num) ? $exif->{ '999999.' . $ifd_num } : $exif;
            if ($parent_tag) {
                $exif_hash->{$parent_tag} = {}
                    if !ref $exif_hash->{$parent_tag};
                $exif_hash->{$parent_tag}->{$tag} = $data_value;
            }
            else {
                $exif_hash->{$tag} = $data_value;
            }
        }
    }
}

# データタイプに応じたデコード
sub read_exif_data ($$$$$) {
    my ( $this, $type, $data, $long_tmpl, $short_tmpl ) = @_;
    if ( $type == 1 ) {
        return unpack( "C", $data );
    }
    elsif ( $type == 2 ) {
        return unpack( "A", $data );
    }
    elsif ( $type == 3 ) {
        return unpack( $short_tmpl, $data );
    }
    elsif ( $type == 4 ) {
        return unpack( $long_tmpl, $data );
    }
    elsif ( $type == 5 ) {
        return join( "/", unpack( "$long_tmpl$long_tmpl", $data ) );
    }
    else {
        return unpack( "H*", $data );
    }
}

sub exif_tag_to_name ($$$;$) {
    my ( $this, $lang, $tag, $parent_tag ) = @_;
    $lang = 'en' if !$lang;
    if ( $parent_tag && $parent_tag == 34853 ) {
        my %GPS_tag = (
            0 => {
                'en' => 'GPS VersionID',
                'ja' => 'GPSタグのバージョン'
            },
            1 => {
                'en' => 'GPS Latitude Ref',
                'ja' => '北緯(N)or南緯(S)'
            },
            2 => { 'en' => 'GPS Latitude', 'ja' => '緯度(数値)' },
            3 => {
                'en' => 'GPS Longitude Ref',
                'ja' => '東経(E)or西経(W)'
            },
            4 => { 'en' => 'GPS Longitude',    'ja' => '経度(数値)' },
            5 => { 'en' => 'GPS Altitude Ref', 'ja' => '高度の基準' },
            6 => { 'en' => 'GPS Altitude',     'ja' => '高度(数値)' },
            7 => {
                'en' => 'GPS Time Stamp',
                'ja' => 'GPS時間(原子時計の時間)'
            },
            8 => {
                'en' => 'GPS Satellites',
                'ja' => '測位に使った衛星信号'
            },
            9 => { 'en' => 'GPS Status', 'ja' => 'GPS受信機の状態' },
            10 =>
                { 'en' => 'GPS Measure Mode', 'ja' => 'GPSの測位方法' },
            11 => { 'en' => 'GPSDOP',        'ja' => '測位の信頼性' },
            12 => { 'en' => 'GPS Speed Ref', 'ja' => '速度の単位' },
            13 => { 'en' => 'GPS Speed',     'ja' => '速度(数値)' },
            14 =>
                { 'en' => 'GPS Track Ref', 'ja' => '進行方向の単位' },
            15 => { 'en' => 'GPS Track', 'ja' => '進行方向(数値)' },
            16 => {
                'en' => 'GPS Img Direction Ref',
                'ja' => '撮影した画像の方向の単位'
            },
            17 => {
                'en' => 'GPS Img Direction',
                'ja' => '撮影した画像の方向(数値)'
            },
            18 => {
                'en' => 'GPS Map Datum',
                'ja' => '測位に用いた地図データ'
            },
            19 => {
                'en' => 'GPS Dest Latitude Ref',
                'ja' => '目的地の北緯(N)or 南緯(S)'
            },
            20 => {
                'en' => 'GPS Dest Latitude',
                'ja' => '目的地の緯度(数値)'
            },
            21 => {
                'en' => 'GPS Dest Longitude Ref',
                'ja' => '目的地の東経(E)or西経(W)'
            },
            22 => {
                'en' => 'GPS Dest Longitude',
                'ja' => '目的地の経度(数値)'
            },
            23 => {
                'en' => 'GPS Dest Bearing Ref',
                'ja' => '目的地の方角の単位'
            },
            24 => {
                'en' => 'GPS Dest Bearing',
                'ja' => '目的の方角(数値)'
            },
            25 => {
                'en' => 'GPS Dest Distance Ref',
                'ja' => '目的地までの距離の単位'
            },
            26 => {
                'en' => 'GPS Dest Distance',
                'ja' => '目的地までの距離(数値)'
            },
            27 => {
                'en' => 'GPS Processing Method',
                'ja' => '測位方式の名称'
            },
            28 => {
                'en' => 'GPS Area Information',
                'ja' => '測位地点の名称'
            },
            29 => { 'en' => 'GPS Date Stamp',   'ja' => 'GPS日付' },
            30 => { 'en' => 'GPS Differential', 'ja' => 'GPS補正測位' },
            31 => {
                'en' => 'GPSH Positioning Error',
                'ja' => '水平方向測位誤差'
            },
        );
        return ( ref $GPS_tag{$tag} ) ? $GPS_tag{$tag}{$lang} : '???';
    }
    else {
        my %tag = (
            256 => { 'en' => 'Image Width',  'ja' => '画像の幅' },
            257 => { 'en' => 'Image Length', 'ja' => '画像の高さ' },
            258 => {
                'en' => 'Bits Per Sample',
                'ja' => '画像のビットの深さ'
            },
            259 => { 'en' => 'Compression', 'ja' => '圧縮の種類' },
            262 => {
                'en' => 'Photometric Interpretation',
                'ja' => '画素構成'
            },
            274 => { 'en' => 'Orientation', 'ja' => '画像方向' },
            277 => {
                'en' => 'Samples Per Pixel',
                'ja' => 'コンポーネント数'
            },
            284 => {
                'en' => 'Planar Configuration',
                'ja' => '画像データの並び'
            },
            530 => {
                'en' => 'Y Cb Cr Sub Sampling',
                'ja' => 'YCC の画素構成(C の間引き率)'
            },
            531 => {
                'en' => 'Y Cb Cr Positioning',
                'ja' => 'YCC の画素構成(Y と C の位置)'
            },
            282 => {
                'en' => 'X Resolution',
                'ja' => '画像の幅の解像度'
            },
            283 => {
                'en' => 'Y Resolution',
                'ja' => '画像の高さの解像度'
            },
            296 => {
                'en' => 'Resolution Unit',
                'ja' => '画像の幅と高さの解像度の単位'
            },
            273 => {
                'en' => 'Strip Offsets',
                'ja' => '画像データのロケーション'
            },
            278 => {
                'en' => 'Rows Per Strip',
                'ja' => '1 ストリップあたりの行の数'
            },
            279 => {
                'en' => 'Strip Byte Counts',
                'ja' => 'ストリップの総バイト数'
            },
            513 => {
                'en' => 'JPEG Interchange Format',
                'ja' => 'JPEG の SOI へのオフセット'
            },
            514 => {
                'en' => 'JPEG Interchange Format Length',
                'ja' => 'JPEG データのバイト数'
            },
            301 => {
                'en' => 'Transfer Function',
                'ja' => '再生階調カーブ特性'
            },
            318 => {
                'en' => 'White Point',
                'ja' => '参照白色点の色度座標値'
            },
            319 => {
                'en' => 'Primary Chromaticities',
                'ja' => '原色の色度座標値'
            },
            529 => {
                'en' => 'Y Cb Cr Coefficients',
                'ja' => '色変換マトリクス係数'
            },
            532 => {
                'en' => 'Reference Black White',
                'ja' => '参照黒色点値と参照白色点値'
            },
            306 =>
                { 'en' => 'Date Time', 'ja' => 'ファイル変更日時' },
            270 =>
                { 'en' => 'Image Description', 'ja' => '画像タイトル' },
            271 => {
                'en' => 'Make',
                'ja' => '画像入力機器のメーカ名'
            },
            272 => {
                'en' => 'Model',
                'ja' => '画像入力機器のモデル名'
            },
            305   => { 'en' => 'Software',  'ja' => 'ソフトウェア' },
            315   => { 'en' => 'Artist',    'ja' => 'アーティスト' },
            33432 => { 'en' => 'Copyright', 'ja' => '著作権' },

            34665 => { 'en' => 'Exif IFD',       'ja' => 'Exif IFD' },
            34853 => { 'en' => 'GPS Infomation', 'ja' => 'GPS 情報' },

            36864 =>
                { 'en' => 'Exif Version', 'ja' => ' Exif バージョン' },
            40960 => {
                'en' => 'Flashpix Version',
                'ja' => '対応フラッシュピックスバージョン'
            },
            40961 => { 'en' => 'Color Space', 'ja' => '色空間情報' },
            42240 => { 'en' => 'Gamma',       'ja' => '再生ガンマ' },
            37121 => {
                'en' => 'Components Configuration',
                'ja' => '各コンポーネントの意味'
            },
            37122 => {
                'en' => 'Compressed Bits Per Pixel',
                'ja' => '画像圧縮モード'
            },
            40962 =>
                { 'en' => 'PixelX Dimension', 'ja' => '実効画像幅' },
            40963 =>
                { 'en' => 'PixelY Dimension', 'ja' => '実効画像高さ' },
            40964 => {
                'en' => 'Related Sound File',
                'ja' => '関連音声ファイル'
            },
            36867 => {
                'en' => 'Date Time Original',
                'ja' => '原画像データの生成日時'
            },
            36868 => {
                'en' => 'Date Time Digitized',
                'ja' => 'デジタルデータの作成日時'
            },
            37520 => {
                'en' => 'Sub Sec Time',
                'ja' => 'ファイル変更日時の小数点以下秒'
            },
            37521 => {
                'en' => 'Sub Sec Time Original',
                'ja' =>
                    '原画像データの作成日時の小数点以下秒'
            },
            37522 => {
                'en' => 'Sub Sec Time Digitized',
                'ja' =>
                    'デジタルデータの作成日時の小数点以下秒'
            },
            42016 => {
                'en' => 'Image UniqueID',
                'ja' => '画像ユニーク ID '
            },
            42032 => {
                'en' => 'Camera Owner Name',
                'ja' => 'カメラ所有者名'
            },
            42033 => {
                'en' => 'Body Serial Number',
                'ja' => 'カメラシリアル番号'
            },
            42034 => {
                'en' => 'Lens Specification',
                'ja' => 'レンズの仕様情報'
            },
            42035 =>
                { 'en' => 'Lens Make', 'ja' => 'レンズのメーカ名' },
            42036 =>
                { 'en' => 'Lens Model', 'ja' => 'レンズのモデル名' },
            42037 => {
                'en' => 'Lens Serial Number',
                'ja' => 'レンズシリアル番号'
            },
            33434 => { 'en' => 'Exposure Time', 'ja' => '露出時間' },
            33437 => { 'en' => 'F Number',      'ja' => 'F値' },
            34850 => {
                'en' => 'Exposure Program',
                'ja' => '露出プログラム'
            },
            34852 => {
                'en' => 'Spectral Sensitivity',
                'ja' => 'スペクトル感度'
            },
            34855 => {
                'en' => 'Photographic Sensitivity',
                'ja' => '撮影感度'
            },
            34856 => { 'en' => 'OECF', 'ja' => '光電変換関数' },
            34864 => { 'en' => 'Sensitivity Type', 'ja' => '感度種別' },
            34865 => {
                'en' => 'Standard Output Sensitivity',
                'ja' => '標準出力感度'
            },
            34866 => {
                'en' => 'Recommended Exposure Index',
                'ja' => '推奨露光指数'
            },

            34867 => { 'en' => 'ISO Speed', 'ja' => 'ISO スピード' },
            34868 => {
                'en' => 'ISO Speed Latitudeyyy',
                'ja' => 'ISO スピードラチチュード yyy'
            },
            34869 => {
                'en' => 'ISO Speed Latitudezzz',
                'ja' => 'ISO スピードラチチュード zzz'
            },
            37377 => {
                'en' => 'Shutter Speed Value',
                'ja' => 'シャッタースピード'
            },
            37378 => { 'en' => 'Aperture Value',   'ja' => '絞り値' },
            37379 => { 'en' => 'Brightness Value', 'ja' => '輝度値' },
            37380 =>
                { 'en' => 'Exposure Bias Value', 'ja' => '露光補正値' },
            37381 => {
                'en' => 'Max Aperture Value',
                'ja' => 'レンズ最小F値'
            },
            37382 =>
                { 'en' => 'Subject Distance', 'ja' => '被写体距離' },
            37383 => { 'en' => 'Metering Mode', 'ja' => '測光方式' },
            37384 => { 'en' => 'Light Source',  'ja' => '光源' },
            37385 => { 'en' => 'Flash',         'ja' => 'フラッシュ' },
            37386 =>
                { 'en' => 'Focal Length', 'ja' => 'レンズ焦点距離' },

            37396 => { 'en' => 'Subject Area', 'ja' => '被写体領域' },
            37500 =>
                { 'en' => 'Maker Note', 'ja' => 'メーカーノート' },
            37510 => {
                'en' => 'User Comment',
                'ja' => 'ユーザーコメント'
            },

            41483 =>
                { 'en' => 'Flash Energy', 'ja' => 'フラッシュ強度' },
            41484 => {
                'en' => 'Spatial Frequency Response',
                'ja' => '空間周波数応答'
            },
            41486 => {
                'en' => 'Focal PlaneX Resolution',
                'ja' => '焦点面の幅の解像度'
            },
            41487 => {
                'en' => 'Focal PlaneY Resolution',
                'ja' => '焦点面の高さの解像度'
            },
            41488 => {
                'en' => 'Focal Plane Resolution Unit',
                'ja' => '焦点面解像度単位'
            },
            41492 =>
                { 'en' => 'Subject Location', 'ja' => '被写体位置' },
            41493 => {
                'en' => 'Exposure Index',
                'ja' => '露出インデックス'
            },
            41495 => { 'en' => 'Sensing Method', 'ja' => 'センサ方式' },
            41728 =>
                { 'en' => 'File Source', 'ja' => 'ファイルソース' },
            41729 => { 'en' => 'Scene Type', 'ja' => 'シーンタイプ' },

            41730 => { 'en' => 'CFA Pattern', 'ja' => '￼CFA パターン' },
            41985 =>
                { 'en' => 'Custom Rendered', 'ja' => '個別画像処理' },
            41986 => { 'en' => 'Exposure Mode', 'ja' => '露出モード' },
            41987 => {
                'en' => 'White Balance',
                'ja' => 'ホワイトバランス'
            },
            41988 => {
                'en' => 'Digital Zoom Ratio',
                'ja' => 'デジタルズーム倍率'
            },
            41989 => {
                'en' => 'Focal Length In35mm Film',
                'ja' => '35mm 換算レンズ焦点距離'
            },
            41990 => {
                'en' => 'Scene Capture Type',
                'ja' => '撮影シーンタイプ'
            },
            41991 => { 'en' => 'Gain Control', 'ja' => 'ゲイン制御' },
            41992 =>
                { 'en' => 'Contrast', 'ja' => '撮影コントラスト' },
            41993 => { 'en' => 'Saturation', 'ja' => '撮影彩度' },
            41994 =>
                { 'en' => 'Sharpness', 'ja' => '撮影シャープネス' },
            41995 => {
                'en' => 'Device Setting Description',
                'ja' => '撮影条件記述情報'
            },
            41996 => {
                'en' => 'Subject Distance Range',
                'ja' => '被写体距離レンジ'
            },

            '999999.1' =>
                { 'en' => 'Thumbnail IFD', 'ja' => 'サムネイル画像' },

        );
        return
              ( ref $tag{$tag} )            ? $tag{$tag}{$lang}
            : ( $tag =~ /^999999\.(\d+)$/ ) ? 'IFD' . $1
            :                                 $tag;
    }
}

sub exif_list ($$;$$) {
    my ( $this, $exif, $lang, $parent_tag ) = @_;
    $lang = 'en' if !$lang;
    my $dl = '<dl>';
    foreach my $tag ( sort { $a <=> $b || $a cmp $b } keys %$exif ) {
        $dl .= '<dt>'
            . $this->exif_tag_to_name( $lang, $tag, $parent_tag ) . '</dt>';
        $dl
            .= '<dd>'
            . ( ( ref $exif->{$tag} )
            ? $this->exif_list( $exif->{$tag}, $lang, $tag )
            : $exif->{$tag} )
            . '</dd>';
    }
    $dl .= '</dl>';
    return $dl;
}

sub latlng_from_exif ($$) {
    my ( $this, $exif ) = @_;
    if ( ref $exif->{'34853'} ) {
        my ( $lat, $lng );
        if ( $exif->{'34853'}{'2'} ) {
            $lat = $this->exifdms_to_deg( $exif->{'34853'}{'2'},
                $exif->{'34853'}{'1'} );
        }
        if ( $exif->{'34853'}{'4'} ) {
            $lng = $this->exifdms_to_deg( $exif->{'34853'}{'4'},
                $exif->{'34853'}{'3'} );
        }
        return ( $lat, $lng );
    }
    return ();
}

sub exifdms_to_deg ($$$) {
    my ( $this, $dms, $ref ) = @_;
    if ( $dms =~ m!(\d+)/(\d+) (\d+)/(\d+) (\d+)/(\d+)! ) {
        my ( $d, $m, $s )
            = ( $1 / ( $2 || 1 ), $3 / ( $4 || 1 ), $5 / ( $6 || 1 ) );
        my $deg = $d + $m / 60 + $s / 3600;
        return ( $ref eq "S" || $ref eq "W" ) ? $deg * -1 : $deg;
    }
    return;
}

######################################################################
#	(2) obj->input_info(input_obj,name,maxwidth,maxheight);
#		アップロードされた画像の情報を取得します。
sub input_info ($$$$;$$$$$$) {
    my ($this,         $input_obj, $name,     $no,
        $maxwidth,     $maxheight, $skip_max, $properties,
        $is_thumbnail, $max_size
    ) = @_;
    my $updated = 0;
    $no += 0;
    if ( !ref $properties ) {
        $properties = {
            $input_obj->get_posted_file_properties_from_param( $name, $no ) };
        $updated = 1;
    }
    elsif ( !$properties->{'type'} ) {
        $updated = 1;
    }
    return () if !%$properties || !$properties->{'filename'};
    my @info = $this->info(
        (   ( $properties->{'savedir'} && $properties->{'savename'} )
            ? $properties->{'savedir'} . '/' . $properties->{'savename'}
            : $input_obj->get_clipboard( $name, $no ) . '/data'
        ),
        $maxwidth,
        $maxheight,
        $skip_max,
        $is_thumbnail,
        $max_size
    );
    return () unless @info >= 4;
    $properties->{'type'}   = $info[3];
    $properties->{'width'}  = $info[0];
    $properties->{'height'} = $info[1];
    $properties->{'filename'} .= '.' . $info[2]
        unless $properties->{'filename'} =~ s/\.[^\.]+$/\.$info[2]/;

    if ( $properties->{'exif'} ) {
        $info[4] = $this->url_decode( $properties->{'exif'} );
    }
    elsif ( !$properties->{'exif'} && ref $info[4] ) {
        $properties->{'exif'} = $input_obj->url_encoded_query( $info[4] );
        $updated = 1;
    }
    $input_obj->set_posted_file_properties_from_param( $name, $no,
        $properties )
        if $updated;
    return @info;
}

sub url_decode ($$;$) {
    my ( $this, $query ) = @_;
    my %data;
    my $hash = {};
    $query =~ s/\+/ /g;
    while ( $query =~ /\G([^=\&]*)=([^\&]*)[\&]?/sgo ) {
        my ( $k, $v ) = ( $1, $2 );
        utf8::encode($k) if utf8::is_utf8($k);
        utf8::encode($v) if utf8::is_utf8($v);
        $k =~ s/%([A-Fa-f0-9][A-Fa-f0-9])/pack("C",hex($1))/ge;
        $v =~ s/%([A-Fa-f0-9][A-Fa-f0-9])/pack("C",hex($1))/ge;
        $data{$k} = $v;

    }
    while ( my ( $k, $v ) = each %data ) {
        my @keys = split( /:/, $k );
        my $temp = $hash;
        for ( my $i = 0; $i < $#keys; $i++ ) {
            $temp->{ $keys[$i] } = {} if !ref $temp->{ $keys[$i] };
            $temp = $temp->{ $keys[$i] };
        }
        $temp->{ $keys[$#keys] } = $v;
    }
    return $hash;
}

#	(2) obj->is_not_a_web_image(input_obj,$name,$no,maxwidth,maxheight,ext,force);
#		アップロードされた画像をすべてチェックします。
sub is_not_a_web_image ($$;$$$$$$) {
    my ($this,      $input_obj, $k,     $no, $maxwidth,
        $maxheight, $ext,       $force, $properties
    ) = @_;
    $no += 0;
    $properties
        = { $input_obj->get_posted_file_properties_from_param( $k, $no ) }
        if !ref $properties;
    if ( !%$properties || !$properties->{'filename'} ) {
        if ($force) {
            push(
                @{ $cgi::action->{'dialog'} },
                $this->decode_utf8(
                    '画像ファイルが不足しています。')
            ) if ref $cgi::action;
            return 1;
        }
        else {
            return 0;
        }
    }
    my @info
        = $this->input_info( $input_obj, $k, $no, $maxwidth, $maxheight, 0,
        $properties );
    if ( !defined $info[2] || !$info[2] ) {
        push( @{ $cgi::action->{'dialog'} }, $this->dialog )
            if ref $cgi::action;
        return 1;
    }
    elsif ( $ext && $info[2] ne $ext ) {
        push(
            @{ $cgi::action->{'dialog'} },
            $this->decode_utf8(
                ( uc $ext ) . '形式ではありません。'
            )
        ) if ref $cgi::action;
        return 1;
    }
    return 0;
}

#	(2) obj->is_not_web_image(input_obj,maxwidth,maxheight,ext,force);
#		アップロードされた画像をすべてチェックします。
sub is_not_web_image ($$;$$$$$) {
    my ( $this, $input_obj, $maxwidth, $maxheight, $ext, $force ) = @_;
    return 0
        unless $input_obj->{'clipboard_id'}
        && opendir( DIR,
        $env::clipboard . '/' . $input_obj->{'clipboard_id'} );
    my @parameters
        = grep( $_ !~ /^\./
            && -d $env::clipboard . '/'
            . $input_obj->{'clipboard_id'} . '/'
            . $_,
        readdir(DIR) );
    closedir(DIR);
PARAM:
    foreach my $k (@parameters) {
        my $clipboard
            = $env::clipboard . '/' . $input_obj->{'clipboard_id'} . '/' . $k;
        next PARAM unless opendir( DIR, $clipboard );
        my @attachments = map( $clipboard . '/' . $_,
            grep( $_ !~ /^\./ && -d $clipboard . '/' . $_, readdir(DIR) ) );
        closedir(DIR);
        my $no = 0;
    ATTACH:
        foreach my $path (@attachments) {
            my %properties = $input_obj->get_posted_file_properties($path);
            return 1
                if $this->is_not_a_web_image( $input_obj,
                $properties{'name'}, $no, $maxwidth, $maxheight, $ext,
                $force, \%properties );
            $no++;
        }
    }
    return 0;
}

######################################################################
#	(3)-1-0 (width,height) = obj->image_resize(o_width,o_height,ratio);
#		指定された割合で拡大縮小した画像サイズを返します。
sub image_resize ($$$$) {
    my ( $this, $width, $height, $ratio ) = @_;
    return ( int( $width * $ratio ), int( $height * $ratio ) );
}

#---------------------------------------------------------------------
#	(3)-1-1 (width,height,o_width,o_height) = obj->resize(image,ratio);
#		指定された割合で拡大縮小した画像サイズを返します。
sub resize ($$$) {
    my ( $this, $image, $ratio ) = @_;
    my @info = $this->info($image);
    return ( $this->image_resize( $info[0], $info[1], $ratio ), @info );
}

#---------------------------------------------------------------------
#	(3)-2-0 (width,height) = obj->image_fit(o_width,o_height,maxwidth,maxheight);
#		指定された最大値を超えないよう縮小した画像サイズを返します。
sub image_fit ($$$$$) {
    my ( $this, $width, $height, $maxwidth, $maxheight ) = @_;
    return (
        ( ( defined $width )  ? $width  : '' ),
        ( ( defined $height ) ? $height : '' )
    ) if !$width || !$height;
    return ( $width, $height )
        if ( !$maxwidth || $width < $maxwidth )
        && ( !$maxheight || $height < $maxheight );
    my $ratio = $width / $height;
    if ( $maxwidth && $width > $maxwidth ) {
        $width  = $maxwidth;
        $height = int( $maxwidth / $ratio );
    }
    if ( $maxheight && $height > $maxheight ) {
        $height = $maxheight;
        $width  = int( $maxheight * $ratio );
    }
    return ( $width, $height );
}

#---------------------------------------------------------------------
#	(3)-2-1 (width,height,o_width,o_height) = obj->fit(image,maxwidth,maxheight);
#		指定された最大値を超えないよう縮小した画像サイズを返します。
sub fit ($$$$) {
    my ( $this, $image, $maxwidth, $maxheight ) = @_;
    my @info = $this->info($image);
    return (
        (          ( !$maxwidth || $info[0] < $maxwidth )
                && ( !$maxheight || $info[1] < $maxheight )
        )
        ? ( $info[0], $info[1], @info )
        : ( $this->image_fit( $info[0], $info[1], $maxwidth, $maxheight ),
            @info
        )
    );
}

######################################################################
#	(4)-0 obj->image_tag(image,o_width,o_height[,border,alt,option,maxwidth,maxheight]);
#		指定された画像の<img />タグを返します。
sub image_tag ($$$$;$$$$$$$$) {
    no warnings qw(numeric);
    my ($this,      $image, $width,       $height,
        $border,    $alt,   $option,      $maxwidth,
        $maxheight, $base,  $base_option, $viewer_disabled
    ) = @_;
    my ( $tag, $open );
    return '' if !$image;
    $alt = $this->escape_for_alt($alt) if $alt;
    $image = $this->base_change( $image, $base, $base_option )
        if $base || $base_option;
    if (   defined $width
        && defined $height
        && (   ( $maxwidth && $width > $maxwidth )
            || ( $maxheight && $height > $maxheight ) )
        )
    {
        my ( $win_width, $win_height ) = ( $width, $height );
        my ( $raw_width, $raw_height ) = ( $width, $height );
        $win_width  = 800 if $win_width > 800;
        $win_height = 600 if $win_height > 600;
        ( $width, $height )
            = $this->image_fit( $width, $height, $maxwidth, $maxheight );
        if ( !$viewer_disabled ) {
            $tag .= '<a href="' . $image . '" target="imageViewer"';
            $tag
                .= " onclick=\"if(typeof cmonos == 'object'){cmonos.window.open('imageViewer',$raw_width,$raw_height)}else{window.open('','imageViewer','width=$win_width,height=$win_height,scrollbars=yes,resizable=yes')} return true\"";
            $tag .= ' title="' . $alt . '"' if defined $alt && $alt ne '';
            $tag .= '>';
            $open = 1;
        }
    }
    $tag .= '<img src="' . $image . '"';
    $tag .= ' width="' . $width . '"' if defined $width && $width;
    $tag .= ' height="' . $height . '"' if defined $height && $height;
    $tag .= ' border="' . $border . '"' if defined $border && $border ne '';
    $tag .= ' alt="' . $alt . '" title="' . $alt . '"'
        if defined $alt && $alt ne '';
    $tag .= ' ' . $option if defined $option && $option;
    $tag .= ' />';
    $tag .= '</a>'        if defined $open   && $open;
    return $this->decode_utf8($tag);
}

sub decode_utf8 ($$) {
    my ( $this, $str ) = @_;
    utf8::decode($str) if defined $str && !ref $str && !utf8::is_utf8($str);
    return $str;
}

#---------------------------------------------------------------------
#	(4)-1 obj->tag(image[,border,alt,option,maxwidth,maxheight]);
#		指定された画像の<img />タグを返します。
sub tag ($$;$$$$$$$) {
    my ($this,     $image,     $border, $alt, $option,
        $maxwidth, $maxheight, $base,   $base_option
    ) = @_;
    if ( $image =~ m!^/! || ref $image ) {
        push(
            @{ $this->{'dialog'} },
            $this->decode_utf8(
                '画像ファイルは相対パスで指定して下さい。'
            )
        );
        return '';
    }
    my ( $width, $height ) = $this->info($image);
    return $this->image_tag(
        $image,  $width,    $height,    $border, $alt,
        $option, $maxwidth, $maxheight, $base,   $base_option
    );
}

#---------------------------------------------------------------------
#	(4)-0 相対パス<->絶対パス(Common.pmから)
sub path_change ($$;$) {
    my ( @dir, $d, $url_d, $base_d );
    my ( $this, $url, $base, $op ) = @_;
    return $url if $base =~ /^\./;
    $base = $ENV{'SCRIPT_NAME'} if !defined $base || $base eq '';
    if ( $url =~ m!^http://! || $url =~ m!^/! ) {
        return $url if $op eq 'abs';
        $url_d  = $1 if $url =~ s!^(http://.*?/|/)!!;
        $base_d = $1 if $base =~ s!^(http://.*?/|/)!!;
        return "$url_d$url"
            if $url_d ne $base_d && $url_d ne '/' && $base_d ne '/';
        @dir = split( '/', $base );
        while ( $d = shift(@dir) ) {
            last unless $url =~ s!^$d/!!;
        }
        foreach ( 0 .. $#dir ) {
            $url = '../' . $url;
        }
        $url = './' . $url if $url !~ /^\.\./;
    }
    else {
        return $url if $op eq 'rel';
        $base =~ s!/[^/]*$!!;
        $d = $url;
        $d =~ s!^\./!!;
        while ( $d =~ s!^\.\./!! ) {
            $base =~ s!/[^/]*$!!;
        }
        $url = $base . '/' . $d;
    }
    return $url;
}

#---------------------------------------------------------------------
# (4)-0 起点の変更
sub base_change ($$$;$) {
    my ( $this, $url, $base, $option ) = @_;
    return $url if !defined $url || $url =~ /:/ || $url =~ /\&#58;/;
    my $here = $ENV{'SCRIPT_NAME'};
    return $this->path_change( $url, $here, 'abs' )
        if defined $option && $option eq 'abs';
    return $url if !defined $base || $base eq '';
    return $this->path_change( $this->path_change( $url, $here, 'abs' ),
        $this->path_change( $base, $here, 'abs' ), 'rel' );
}

######################################################################
#	(5) obj->gif_cat($@images,out,vector);
#		GIF画像を結合します。
sub gif_cat ($$$$$@) {
    my ( $this, $images, $gif, $vector, $type, @colorsetting ) = @_;
    my (@images,                    $handle,
        $logical_screen_descriptor, $r,
        $c,                         $max,
        $c_pos,                     $r_pos,
        $image_height,              $image_width,
        @left_position,             @top_position,
        @image_file,                $pos,
        $buf,                       $flag
    );
    local ( *OUTPUT, *IMAGE );
    @images = @$images;

    # 縦方向の場合の処理（縦書き配列に変換）
    if ( $vector eq 'v' ) {
        my ( @temp, @temp2 );
        foreach (@images) {
            if ( $_ =~ /^[\n\r]$/ ) {
                unshift( @temp2, "\n", @temp );
                undef @temp;
            }
            else {
                push( @temp, $_ );
            }
        }
        unshift( @temp2, @temp );
        @images = @temp2;
        undef @temp;
        undef @temp2;
    }

    #色変換指定がある場合
    my ( @RGB, $option );
    if (@colorsetting) {

        #引数の調整
        $this->color_setting( $type, \@colorsetting, \@RGB, \$option );
    }

    # GIFの論理画面記述部を調べる
    ( $r, $c, $max, $c_pos, $r_pos, $image_height, $image_width )
        = ( 0, 0, 0, 0, 0, 0, 0 );
    foreach (@images) {

        # 改列処理
        if ( $_ =~ /^[\n\r]$/ ) {
            if ( $vector eq 'v' ) {
                $image_height = $c_pos if $image_height < $c_pos;
            }
            else {
                $image_width = $c_pos if $image_width < $c_pos;
            }
            $r_pos += $max;
            $c = $c_pos = $max = 0;
            $r++;
            next;
        }

        # GIFかどうかチェック/縦横サイズの取得
        my ( $image, $width, $height );
        $image = $this->open_image($_);
        next if !defined $image || $image eq '';
        if ( $this->read_image( $image, 0, 3 ) ne "\x47\x49\x46" ) {
            $this->close_image($image);
            next;
        }

        $width  = unpack( 'v', $this->read_image( $image, 6, 2 ) );
        $height = unpack( 'v', $this->read_image( $image, 8, 2 ) );
        if ( $width == 0 || $height == 0 ) {
            $this->close_image($image);
            next;
        }

        # 詰め込み(1) + 背景色(1) + 縦横比(1) + 広域配色表
        if ( !defined $logical_screen_descriptor
            || $logical_screen_descriptor eq '' )
        {
            ($buf) = $this->color_table_size( $image, 10 );

            #色変換
            if ( $type && @colorsetting ) {
                $logical_screen_descriptor
                    = $this->read_image( $image, 10, 3 );
                $logical_screen_descriptor
                    .= $this->change_color(
                    $this->read_image( $image, 13, $buf ),
                    $type, \@RGB, $option );
            }
            else {
                $logical_screen_descriptor
                    = $this->read_image( $image, 10, 3 + $buf );
            }
        }
        $this->close_image($image);

        # 方向別に処理
        if ( $vector eq 'v' ) {
            $max = $width if $width > $max;
            $left_position[$r][$c] = $r_pos;
            $top_position[$r][$c]  = $c_pos;
            $c_pos += $height;
        }
        else {
            $max = $height if $height > $max;
            $left_position[$r][$c] = $c_pos;
            $top_position[$r][$c]  = $r_pos;
            $c_pos += $width;
        }
        $image_file[$r][$c] = $_;

        $c++;
    }
    if ( $vector eq 'v' ) {
        $image_height = $c_pos if $image_height < $c_pos;
        $image_width = $r_pos + $max;
    }
    else {
        $image_width = $c_pos if $image_width < $c_pos;
        $image_height = $r_pos + $max;
    }

    if ( $image_width == 0 || $image_height == 0 ) {
        push(
            @{ $this->{'dialog'} },
            $this->decode_utf8('GIF画像の結合に失敗しました。')
        );
        return;
    }

    # 出力先を開く
    $handle = $this->open_out($gif);

    # ヘッダと論理画面記述部を出力
    print $handle "GIF89a";
    print $handle pack( "v", $image_width ) . pack( "v", $image_height );
    print $handle $logical_screen_descriptor;

ROW:
    for ( $r = 0; $r <= $#image_file; $r++ ) {
    IMAGE:
        for ( $c = 0; $c <= $#{ $image_file[$r] }; $c++ ) {
            my ( @color_table_info, $image, $label );
            $image = $this->open_image( $image_file[$r][$c] );

            # 論理画面記述部をスキップ
            $pos = 10;

            # 広域配色表の記録
            $this->color_setting( $type, \@colorsetting, \@RGB, \$option )
                if grep( $_ eq 'RANDOM', @colorsetting );
            @color_table_info = $this->color_table_size( $image, $pos, 3 );
            $color_table_info[3] = $this->change_color( $color_table_info[3],
                $type, \@RGB, $option )
                if $type && @colorsetting;

# 詰め込み(1) + 背景色(1) + 縦横比(1) + 広域配色表　をスキップ
            $pos += $color_table_info[0] + 3;

        LABEL:
            while ( $label = $this->read_image( $image, $pos, 1 ) ) {
                if ( $label eq "\x3b" || bytes::length $label == 0 ) {
                    $this->close_image($image);
                    next IMAGE;
                }

                print $handle $label;
                $pos++;

                #画像記述部
                if ( $label eq "\x2c" ) {
                    print $handle pack( "v",
                        unpack( 'v', $this->read_image( $image, $pos, 2 ) )
                            + $left_position[$r][$c] );
                    $pos += 2;
                    print $handle pack( "v",
                        unpack( 'v', $this->read_image( $image, $pos, 2 ) )
                            + $top_position[$r][$c] );
                    $pos += 2;

                    #画像幅/高さ
                    print $handle $this->read_image( $image, $pos, 4 );
                    $pos += 4;

# 詰め込み(1) + 狭域配色表の出力（広域配色表を狭域配色表として出力）
                    ( $buf, $flag ) = $this->color_table_size( $image, $pos );
                    if ( $flag =~ /^1/ ) {

                        #色変換
                        if ( $type && @colorsetting ) {
                            print $handle $this->read_image( $image, $pos,
                                1 );
                            print $handle $this->change_color(
                                $this->read_image( $image, $pos + 1, $buf ),
                                $type, \@RGB, $option );
                        }
                        else {
                            print $handle $this->read_image( $image, $pos,
                                1 + $buf );
                        }
                        $pos += 1 + $buf;
                    }
                    elsif ( $color_table_info[0] > 0 ) {
                        print $handle pack( "B8",
                                  '1'
                                . bytes::substr( $flag, 1, 3 )
                                . $color_table_info[2] );
                        print $handle $color_table_info[3];
                        $pos++;
                    }
                    else {
                        print $handle $this->read_image( $image, $pos, 1 );
                        $pos++;
                    }

                    # LZW最小符号寸法(1)
                    print $handle $this->read_image( $image, $pos, 1 );
                    $pos++;

                    # 画像データ（副ブロック）
                    $pos = $this->print_sub_blocks( $handle, $image, $pos );

                    next LABEL;

                    #拡張部
                }
                elsif ( $label eq "\x21" ) {

                    #ラベル
                    my $xlabel = $this->read_image( $image, $pos, 1 );
                    $pos++;
                    print $handle $xlabel;

                    #グラフィック機能拡張
                    if ( $xlabel eq "\xF9" ) {
                        my ($method);

                        #ブロック寸法
                        print $handle "\x04";
                        $pos++;

                        #詰め込み欄
                        $method = unpack( "B8",
                            $this->read_image( $image, $pos, 1 ) );
                        bytes::substr( $method, 3, 3, '001' );
                        print $handle pack( "B8", $method );
                        $pos++;

                        #それ以降
                        print $handle $this->read_image( $image, $pos, 4 );
                        $pos += 4;
                        next LABEL;
                    }

                    # 拡張データ（副ブロック）
                    $pos = $this->print_sub_blocks( $handle, $image, $pos );
                    next LABEL;
                }
            }
            $this->close_image($image);
        }
    }

    #終了部
    print $handle "\x3b";

    close($handle);
    $this->clear_http_header;
    return ( $image_width, $image_height );
}

#---------------------------------------------------------------------
#	(5)-0-1 obj->color_table_size(image,pos,jump);
#		配色表のサイズを返します。
sub color_table_size ($$$;$) {
    my ( $this, $image, $pos, $jump ) = @_;
    my ( $flag, $buf, $info, $table );
    $flag = unpack( "B8", $this->read_image( $image, $pos, 1 ) );
    if ( bytes::substr( $flag, 0, 1 ) ) {
        $info = bytes::substr( $flag, 4, 4 );
        $buf = (
            2**(unpack( "n",
                    pack( "B16", "0" x 13 . bytes::substr( $flag, 5, 3 ) ) )
                    + 1
            )
        ) * 3;
    }
    $table = $this->read_image( $image, $pos + $jump, $buf )
        if defined $jump && $jump > 0;
    return ( $buf, $flag, $info, $table );
}

#---------------------------------------------------------------------
#	(5)-0-2 obj->print_sub_blocks(handle,image,pos);
#		副ブロックを出力します。
sub print_sub_blocks ($$$) {
    my ( $this, $handle, $image, $pos ) = @_;
    my ( $buf, $buf_temp );
    $buf = 0;
    while ( $buf_temp
        = unpack( "C", $this->read_image( $image, $pos + $buf, 1 ) ) )
    {
        $buf += $buf_temp + 1;
    }
    print $handle $this->read_image( $image, $pos, $buf );
    return $pos + $buf;
}

#---------------------------------------------------------------------
#	(5)-0-3 出力先を開く
sub open_out ($*) {
    my ( $this, $out ) = @_;
    if ( ref $out eq 'GLOB' ) {
        binmode $$out;
        return $$out;
    }
    elsif ( $out ne '' && open( OUTPUT, ">>$out" ) ) {
        binmode OUTPUT;
        flock( OUTPUT, LOCK_EX );
        truncate( OUTPUT, 0 );
        seek( OUTPUT, 0, 0 );
        return \*OUTPUT;
    }
    else {
        binmode STDOUT;
        print STDOUT 'Content-Type: image/gif' . "\n";
        print STDOUT $this->{'HTTP_HEADER'}
            if defined $this->{'HTTP_HEADER'} && $this->{'HTTP_HEADER'} ne '';
        print STDOUT $this->{'Set_Cookie'}
            if defined $this->{'Set_Cookie'} && $this->{'Set_Cookie'} ne '';
        print STDOUT "\n";
        return \*STDOUT;
    }
}

#---------------------------------------------------------------------
#	(5)-0 obj->set_cookie(limit,cname,cookie,domain,path);
#		Cookieをセットします。
sub set_cookie ($$$$$$) {
    my ( $this, $lifetime, $cname, $cookie, $domain, $path ) = @_;
    my ($expires);
    return '' if !$cname || !defined $cookie || $cookie eq '';
    $path = ' path=' . ( ( $path ne '' ) ? $path : '/' ) . ';';
    $expires = ' expires=' . $this->get_expires($lifetime) . ';' if $lifetime;
    if ( defined $domain && $domain ne '' ) {
        $domain = lc $1 if $domain =~ m!^https?://([^/]+)!i;
        $domain = ' domain=' . $domain . ';'
            if defined $domain && $domain ne '';
    }
    $this->{'Set_Cookie'}
        .= 'Set-Cookie: '
        . $cname . '='
        . $cookie . ';'
        . $expires
        . $path
        . $domain
        . (
        (   $ENV{'HTTPS'}
                || ( $ENV{'SERVER_PORT'} && $ENV{'SERVER_PORT'} == 443 )
        ) ? 'secure' : ''
        ) . "\x0A";
}

sub get_expires ($$) {
    my ( $this, $lifetime ) = @_;
    my ( $time, @week, @month, @gmt );
    $time = time + int( $lifetime * 3600 );
    @week = (
        'Sunday',   'Monday', 'Tuesday', 'Wednesday',
        'Thursday', 'Friday', 'Saturday'
    );
    @month = (
        'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
        'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'
    );
    @gmt = gmtime($time);
    $gmt[0] = sprintf( "%.2d", $gmt[0] );
    $gmt[1] = sprintf( "%.2d", $gmt[1] );
    $gmt[2] = sprintf( "%.2d", $gmt[2] );
    $gmt[3] = sprintf( "%.2d", $gmt[3] );
    $gmt[5] += 1900 if $gmt[5] < 1900;
    return
        "$week[$gmt[6]], $gmt[3]-$month[$gmt[4]]-$gmt[5] $gmt[2]:$gmt[1]:$gmt[0] GMT";
}

#---------------------------------------------------------------------
#	(5)-0 obj->set_http_header(key,value);
#		HTTPヘッダをセットします。
sub set_http_header ($@) {
    my ( $this, @headers ) = @_;
    for ( my $i = 0; $i < @headers; $i += 2 ) {
        $this->{'HTTP_HEADER'} .= $this->decode_utf8( $headers[$i] ) . ':'
            . $this->decode_utf8( $headers[ $i + 1 ] ) . "\x0A";
    }
}

sub clear_http_header ($) {
    my $this = shift;
    delete $this->{'HTTP_HEADER'};
    delete $this->{'Set_Cookie'};
}

######################################################################
#	(6) obj->change_color_table(image,out,type,color[,option[,rev]]);
#		GIF画像の配色表を変換します。
sub change_color_table ($$$$;@) {
    my ( $this, $gif, $out, $type, @setting ) = @_;
    my ($image,     $handle, $pos,  $buf, $width, $height,
        $same_flag, $label,  @mode, @RGB, $option
    );
    local ( *OUTPUT, *IMAGE );

    # GIFかどうかチェック
    $image = $this->open_image($gif);
    if ( !defined $image || $image eq '' ) {
        push(
            @{ $this->{'dialog'} },
            $this->decode_utf8('画像がみつかりません。')
        );
        return;
    }
    elsif ( $this->read_image( $image, 0, 3 ) ne "\x47\x49\x46" ) {
        $this->close_image($image);
        push(
            @{ $this->{'dialog'} },
            $this->decode_utf8('GIF画像ではありません。')
        );
        return;
    }

    #同じファイル？
    if ( $gif eq $out ) {
        $same_flag = 1;
        $out .= '.temp';
    }

    $width  = unpack( 'v', $this->read_image( $image, 6, 2 ) );
    $height = unpack( 'v', $this->read_image( $image, 8, 2 ) );

    #引数の調整
    $this->color_setting( $type, \@setting, \@RGB, \$option );

    # 出力先を開く
    $handle = $this->open_out($out);

    $pos = 0;
    print $handle $this->read_image( $image, $pos, 13 );
    $pos += 13;

    # 詰め込み(1) + 背景色(1) + 縦横比(1) + 広域配色表
    ($buf) = $this->color_table_size( $image, 10 );
    print $handle $this->change_color(
        $this->read_image( $image, $pos, $buf ),
        $type, \@RGB, $option )
        if $buf > 0;
    $pos += $buf;

    while ( $label = $this->read_image( $image, $pos, 1 ) ) {
        if ( $label eq "\x3b" || bytes::length $label == 0 ) {
            $this->close_image($image);
            last;
        }

        print $handle $label;
        $pos++;

        #画像記述部
        if ( $label eq "\x2c" ) {

            #画像幅/高さ
            print $handle $this->read_image( $image, $pos, 8 );
            $pos += 8;

            # 狭域配色表サイズ取得
            ($buf) = $this->color_table_size( $image, $pos );

            # 詰め込み(1)の出力
            print $handle $this->read_image( $image, $pos, 1 );
            $pos++;

            # 狭域配色表の出力
            if ( defined $buf && $buf > 0 ) {
                print $handle $this->change_color(
                    $this->read_image( $image, $pos, $buf ),
                    $type, \@RGB, $option );
                $pos += $buf;
            }

            # LZW最小符号寸法(1)
            print $handle $this->read_image( $image, $pos, 1 );
            $pos++;

            # 画像データ（副ブロック）
            $pos = $this->print_sub_blocks( $handle, $image, $pos );

            next;

            #拡張部
        }
        elsif ( $label eq "\x21" ) {

            #ラベル
            print $handle $this->read_image( $image, $pos, 1 );
            $pos++;

            # 拡張データ（副ブロック）
            $pos = $this->print_sub_blocks( $handle, $image, $pos );

            next;
        }
    }

    #オリジナル画像を閉じる
    $this->close_image($image);

    #終了部
    print $handle "\x3b";

    close($handle);
    $this->clear_http_header;

    if ( defined $same_flag && $same_flag ) {
        rename( $out, $gif );
    }
    return ( $width, $height );
}

#引数の調整
sub color_setting ($$$$) {
    my ( $this, $type, $setting, $RGB, $option ) = @_;
    if ( $type eq 'slide' ) {
        if ( $$setting[0] =~ /^random$/i ) {
            $RGB->[0]
                = [ int( rand(256) ), int( rand(256) ), int( rand(256) ),
                '+' ];
        }
        elsif ( $$setting[0]
            =~ /^([\+\-]*)[#]*([a-fA-F0-9]{2})([a-fA-F0-9]{2})([a-fA-F0-9]{2})/
            )
        {
            $RGB->[0] = [ hex($2), hex($3), hex($4), $1 ];
        }
        if ( $$setting[2] =~ /^random$/i ) {
            $RGB->[1]
                = [ int( rand(256) ), int( rand(256) ), int( rand(256) ),
                '-' ];
        }
        elsif ( $$setting[2]
            =~ /^([\+\-]*)[#]*([a-fA-F0-9]{2})([a-fA-F0-9]{2})([a-fA-F0-9]{2})/
            )
        {
            $RGB->[1] = [ hex($2), hex($3), hex($4), $1 ];
        }
        $$option = $$setting[1];
    }
    else {
        @$setting = split( /[\x09\x0A\x0C\x0D\x20]*,[\x09\x0A\x0C\x0D\x20]*/,
            $$setting[0] )
            if @$setting == 1 && $$setting[0] =~ /,/;
        foreach ( 0 .. $#{$setting} ) {
            if ( $$setting[$_] =~ /^random$/i ) {
                $RGB->[$_] = 'random';
            }
            elsif ( $$setting[$_] =~ /^random$/i ) {
                $RGB->[$_] = [ int( rand(256) ), int( rand(256) ),
                    int( rand(256) ) ];
            }
            elsif ( $$setting[$_]
                =~ /^[#]*([a-fA-F0-9]{2})([a-fA-F0-9]{2})([a-fA-F0-9]{2})/ )
            {
                $RGB->[$_] = [ hex($1), hex($2), hex($3) ];
            }
        }
        $$option = @$setting;
    }
}

#---------------------------------------------------------------------
#	(6)-0 obj->change_color(colortable,type,\@RGB,$option);
#		配色表の色を変換します。
sub change_color ($$$$$) {
    my ( $this, $colortable, $type, $RGB, $option ) = @_;
    my ( $newtable, $no, $pos, $r, $g, $b );

    $no  = 0;
    $pos = 0;
    while ( ( $r, $g, $b )
        = unpack( "CCC", bytes::substr( $colortable, $pos, 3 ) ) )
    {
        last if bytes::length $b == 0;
        $pos += 3;

        if ( $type eq 'slide' ) {
            if ( defined $RGB->[0]->[3] && $RGB->[0]->[3] eq '-' ) {
                $r -= $RGB->[0]->[0] if $RGB->[0]->[0] > 0;
                $g -= $RGB->[0]->[1] if $RGB->[0]->[1] > 0;
                $b -= $RGB->[0]->[2] if $RGB->[0]->[2] > 0;
            }
            else {
                $r += $RGB->[0]->[0] if $RGB->[0]->[0] > 0;
                $g += $RGB->[0]->[1] if $RGB->[0]->[1] > 0;
                $b += $RGB->[0]->[2] if $RGB->[0]->[2] > 0;
            }
            if ( defined $RGB->[1]->[3] && $RGB->[1]->[3] eq '-' ) {
                $r -= $RGB->[1]->[0] if $RGB->[1]->[0] > 0;
                $g -= $RGB->[1]->[1] if $RGB->[1]->[1] > 0;
                $b -= $RGB->[1]->[2] if $RGB->[1]->[2] > 0;
            }
            else {
                $r += $RGB->[1]->[0] if $RGB->[1]->[0] > 0;
                $g += $RGB->[1]->[1] if $RGB->[1]->[1] > 0;
                $b += $RGB->[1]->[2] if $RGB->[1]->[2] > 0;
            }
            if ( defined $option && $option ) {
                $r = 510 - $r if $r > 255;
                $g = 510 - $g if $g > 255;
                $b = 510 - $b if $b > 255;
                $r = abs $r   if $r < 0;
                $g = abs $g   if $g < 0;
                $b = abs $b   if $b < 0;
            }
            else {
                $r = 255 if $r > 255;
                $g = 255 if $g > 255;
                $b = 255 if $b > 255;
                $r = 0   if $r < 0;
                $g = 0   if $g < 0;
                $b = 0   if $b < 0;
            }

        }
        else {
            my $rgb = $RGB->[ $no % $option ];
            ( $r, $g, $b )
                = ( ref $rgb )
                ? @$rgb
                : ( int( rand(256) ), int( rand(256) ), int( rand(256) ) );
            $no++;
        }

        $newtable .= pack( "C", $r ) . pack( "C", $g ) . pack( "C", $b );
    }
    return $newtable;
}

#---------------------------------------------------------------------
#	(6)-3 ディレクトリの中にあるGIF画像を一括変換する
sub change_color_table_multi ($$$$;@) {
    my ( $this, $dir, $out, $type, @setting ) = @_;
    my ( @old, @images, $same_flag, $check );

    #ディレクトリが存在しないとき
    return unless -d $dir;

    #書き出し先ディレクトリを作成
    return
           unless -d $out
        || mkdir( $out, $env::dir_mode )
        || ( eval 'use File::Path'
        && eval { mkpath( $out, 0, $env::dir_mode ) } );

    #書き出し先ディレクトリをクリア
    local (*IMAGEDIR);
    if ( $dir eq $out ) {
        $out .= '/' if $out && $out !~ m!/$!;
        $same_flag = 1;
    }
    else {
        opendir( IMAGEDIR, $out );
        $out .= '/' if $out && $out !~ m!/$!;
        @old = grep { $_ =~ /\.(gif|png|jpg|jpeg)$/i && -f $out . $_ }
            readdir(IMAGEDIR);
        closedir(IMAGEDIR);
        if (@old) {
            map( $_ = $out . $_, @old );
            unlink @old;
        }
    }

    $check = 1;
    opendir( IMAGEDIR, $dir );
    $dir .= '/' if $dir && $dir !~ m!/$!;
    @images = grep { $_ !~ /^\./ && -f $dir . $_ } readdir(IMAGEDIR);
    close(IMAGEDIR);

    #配色表スライド値調整
    $setting[0] = '+' . $this->rand_color
        if $type eq 'slide' && $setting[0] eq 'random';

    foreach (@images) {
        my ( $file, $copy ) = ( $dir . $_, $out . $_ );
        local ( *IMG, *COPY );

        #GIF画像変換
        if ( $_ =~ /\.gif$/i ) {
            $check = 0
                unless $this->change_color_table( $file, $copy, $type,
                @setting );

            #それ以外のファイルをコピー
        }
        elsif ( !$same_flag && open( IMG, $file ) ) {
            binmode IMG;
            flock( IMG, LOCK_EX );
            if ( open( COPY, ">>$copy" ) ) {
                binmode COPY;
                flock( COPY, LOCK_EX );
                truncate( COPY, 0 );
                seek( COPY, 0, 0 );
                while (<IMG>) {
                    print COPY;
                }
                close(COPY);
            }
            close(IMG);
            chmod( $env::file_mode, $copy );

        }
        else {
            $check = 0;

        }
    }

    chmod( $env::dir_mode, $out );
    return $check;
}

sub rand_color ($) {
    my $this = shift;
    my @h = ( 0 .. 9, 'a' .. 'f' );
    return
          '#'
        . $h[ int( rand(16) ) ]
        . $h[ int( rand(16) ) ]
        . $h[ int( rand(16) ) ]
        . $h[ int( rand(16) ) ]
        . $h[ int( rand(16) ) ]
        . $h[ int( rand(16) ) ];
}

sub escape_for_alt ($$) {
    use utf8;
    my ( $this, $text ) = @_;
    utf8::decode($text);
    $text
        =~ s/\&(?!(?:#[0-9]{2,6}|#x[a-fA-F0-9]+|[a-zA-Z0-9]{2,10});)/\&amp;/g;
    $text =~ s/(\x0D\x0A|\x0A|\x0D)/ /gs;
    $text =~ s/"/&quot;/g;
    $text =~ s/</&lt;/g;
    $text =~ s/>/&gt;/g;
    $text =~ s/[\x0A\x0D]//sg;
    return $this->decode_utf8($text);
}

########################################################################
#メッセージ
sub dialog ($) {
    my $this   = shift;
    my @dialog = @{ $this->{'dialog'} };
    undef @{ $this->{'dialog'} };
    return @dialog;
}

######################################################################
#sub import {
BEGIN {
    $env::file_mode = 0644 if !defined $env::file_mode || !$env::file_mode;
    $env::dir_mode  = 0755 if !defined $env::dir_mode  || !$env::dir_mode;
    $env::cgi_file_mode = 0755
        if !defined $env::cgi_file_mode || !$env::cgi_file_mode;
}

1;
